import json
from urllib.parse import quote_plus
from urllib.request import Request, urlopen


def post_worker_data_to_monitor(address, key, post_dict):
    try:
        request = Request(f'{address}/{key}/worker_update/', quote_plus(json.dumps(post_dict)).encode())
        # print(f'{address}/{key}/worker_update/')
        urlopen(request)
    except:
        pass


STATUS_CHOICES = {
    'start': 'Starting',
    'halt': 'Halting',
    'stop': 'Stopped',
    'run': 'Running',
    'error': 'Error',
    'wait': 'Waiting',
    'queued': 'Queued',
}


class Status:
    def __init__(self, value: str = 'stop'):
        self.value = None
        self.set(value)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.verbose

    def set(self, value):
        if value in STATUS_CHOICES:
            self.value = value
        else:
            raise TypeError(f'{value} is and invalid status choice. Valid choices are {STATUS_CHOICES.keys()}')

    @property
    def verbose(self):
        return STATUS_CHOICES[self.value]


import ast
from datetime import datetime, timedelta
import calendar


class Cron:
    def __init__(
            self,
            value: str,
            utc_offset: int = 0,
    ):
        self.value = value
        self.utc_offset = utc_offset

        cron_list = self.value.split(' ')

        if len(cron_list) != 5:
            raise ValueError(f'Cron string {self.value} is not the correct format. Should be 5 elements in a string "* * * * *"')

        self.minute = CronValue(cron_list[0])
        self.hour = CronValue(cron_list[1])
        self.day_of_month = CronValue(cron_list[2])
        self.month = CronValue(cron_list[3])
        self.day_of_week = CronValue(cron_list[4])

        self.next_datetime = None

        self.set_next_datetime()

    def is_past_next_datetime(self):
        if datetime.now().replace(second=0, microsecond=0) >= self.next_datetime:
            self.set_next_datetime()
            return True
        else:
            return False

    def set_next_datetime(self):
        ndt = datetime.utcnow().replace(second=0, microsecond=0) + timedelta(hours=self.utc_offset)
        now = datetime.utcnow().replace(second=0, microsecond=0) + timedelta(hours=self.utc_offset)

        # Minute

        if self.minute.function == 'every':
            ndt += timedelta(minutes=1)
        elif self.minute.function == 'specific':
            ndt = ndt.replace(minute=self.minute.specific)

            if now.minute >= ndt.minute:
                ndt += timedelta(hours=1)
        elif self.minute.function == 'step':
            count_step_list = [0]
            count = self.minute.step
            while count < 60:
                count_step_list.append(count)
                count += self.minute.step

            for count_step in count_step_list:
                if count_step > now.minute:
                    ndt = ndt.replace(minute=count_step)
                    break
            else:
                ndt = ndt.replace(minute=0)
                ndt += timedelta(hours=1)

        # Hour

        if self.hour.function == 'every':
            pass

        if self.hour.function == 'specific':
            ndt = ndt.replace(hour=self.hour.specific)

            if ndt.hour == self.hour.specific:               
                if self.minute.function == 'specific':
                    if now.minute >= ndt.minute:
                        ndt += timedelta(days=1)

            if ndt.hour > self.hour.specific:
                ndt += timedelta(days=1)

        elif self.hour.function == 'step':
            count_step_list = [0]
            count = self.hour.step
            while count < 24:
                count_step_list.append(count)
                count += self.hour.step

            for count_step in count_step_list:
                if count_step > now.hour:
                    ndt = ndt.replace(hour=count_step)
                    break
            else:
                ndt = ndt.replace(hour=0)
                ndt += timedelta(days=1)

        # Day of Month

        if self.day_of_month.function == 'every':
            pass

        elif self.day_of_month.function == 'specific':
            ndt = ndt.replace(day=self.day_of_month.specific)

            if ndt.month == 12:
                next_month = 1
            else:
                next_month = ndt.month + 1

            if ndt.day == self.day_of_month.specific:

                if self.hour.function == 'specific':
                    if now.hour >= ndt.hour:
                        ndt = ndt.replace(month=next_month)

            if ndt.day > self.day_of_month.specific:
                ndt = ndt.replace(month=next_month)

        # Month

        if self.month.function == 'every':
            pass

        elif self.month.function == 'specific':
            ndt = ndt.replace(month=self.month.specific)

            if ndt.month == self.month.specific:

                if self.day_of_month.function == 'specific':
                    if now.day >= ndt.day:
                        ndt = ndt.replace(year=(ndt.year + 1))

            if now.month > ndt.month:
                ndt = ndt.replace(year=(ndt.year + 1))

        # Day of Week

        if self.day_of_week.function == 'every':
            pass

        if self.day_of_week.function == 'specific':
            if self.day_of_week.specific == 0:
                self.day_of_week.specific = 7

            if ndt.isoweekday() == self.day_of_week.specific:
                pass
                # print(f'{now.isoweekday() = } {ndt.isoweekday() = } {self.day_of_week.specific = }')
                # if self.hour.function == 'specific':
                #     if now.hour >= ndt.hour:
                #         ndt += timedelta(days=7)

            elif ndt.isoweekday() > self.day_of_week.specific:
                ndt += timedelta(days=(7 - (ndt.isoweekday() - self.day_of_week.specific)))

            else:
                ndt += timedelta(days=(self.day_of_week.specific - ndt.isoweekday()))

        self.next_datetime = ndt


class CronValue:
    def __init__(self, value: str, ):
        self.value = value

        FUNCTION_CHOICES = ('every', 'specific', 'range', 'step')

        self.function = None

        self.specific = None

        self.range_start = None
        self.range_stop = None

        self.step_start = None
        self.step = None

        self.process()

    def process(self):
        range_list = self.value.split('-')

        value_error = 'Invalid cron string used.'

        if len(range_list) == 2:
            self.function = 'range'
            self.range_start = int(range_list[0])
            self.range_stop = int(range_list[1])
        elif len(step_list := self.value.split('/')) == 2:
            if step_list[0] == '*':
                self.function = 'step'
                self.step_start = step_list[0]
                self.step = int(step_list[1])
            else:
                raise ValueError(value_error)
        else:
            if range_list[0] == '*':
                self.function = 'every'
            elif isinstance(ast.literal_eval(range_list[0]), int):
                self.function = 'specific'
                self.specific = int(range_list[0])
            else:
                raise ValueError(value_error)






from datetime import datetime, timedelta

from robit.core.cron import Cron

CREATED_DATE_FORMAT = '%b %d, %Y %I:%M%p'


class Clock:
    def __init__(
            self,
            cron: str = '* * * * *',
            utc_offset: int = 0,
            timer_duration_list_max: int = 10
    ):
        self.cron = Cron(value=cron, utc_offset=utc_offset)

        self.utc_offset = utc_offset

        self.created_utc = datetime.utcnow()
        self.created_tz = datetime.utcnow() + timedelta(hours=utc_offset)

        self.timer_duration_list_max = timer_duration_list_max
        self.timer_duration_list = []
        self.timer_average_duration = 0.0
        self.timer_last_duration = 0.0
        self.timer_shortest_duration = 0.0
        self.timer_longest_duration = 0.0
        self.timer = None

    def as_dict(self):
        return {
            'created': self.created_tz_verbose,
            'now': self.now_tz_verbose,
            'next_run_datetime': self.next_run_datetime_verbose,
            'timer_last_duration': f'{self.timer_last_duration:.2f}',
            'timer_average_duration': f'{self.timer_average_duration:.2f}',
            'timer_shortest_duration': f'{self.timer_shortest_duration:.2f}',
            'timer_longest_duration': f'{self.timer_longest_duration:.2f}',
        }

    def calculate_timer_average_duration(self):
        timer_total_duration = 0.0
        if len(self.timer_duration_list) > 0:
            for timer_duration in self.timer_duration_list:
                timer_total_duration += timer_duration
            self.timer_average_duration = timer_total_duration / len(self.timer_duration_list)

    @property
    def created_utc_verbose(self):
        return self.created_utc.strftime(CREATED_DATE_FORMAT)

    @property
    def created_tz_verbose(self):
        return self.created_tz.strftime(CREATED_DATE_FORMAT)

    @property
    def now_tz_verbose(self):
        return (datetime.utcnow() + timedelta(hours=self.utc_offset)).strftime(CREATED_DATE_FORMAT)

    def is_past_next_run_datetime(self):
        if self.cron.is_past_next_datetime():
            return True
        else:
            return False

    @property
    def next_run_datetime_verbose(self):
        return self.cron.next_datetime.strftime(CREATED_DATE_FORMAT)

    def start_timer(self):
        self.timer = datetime.utcnow()

    def stop_timer(self):
        duration = (datetime.utcnow() - self.timer).total_seconds()

        self.timer_last_duration = duration
        self.timer_duration_list.insert(0, duration)

        if self.timer_shortest_duration == 0 or duration < self.timer_shortest_duration:
            self.timer_shortest_duration = duration

        if self.timer_longest_duration < duration:
            self.timer_longest_duration = duration

        self.calculate_timer_average_duration()
        self.trim_timer_duration_list()

    def trim_timer_duration_list(self):
        if len(self.timer_duration_list) > self.timer_duration_list_max:
            del self.timer_duration_list[self.timer_duration_list_max:]





import json
from http.server import HTTPServer

from robit.core.web_server import WebServer, WebRequestHandler, html_encode_file


class WorkerWebServer(WebServer):
    def httpd_serve(self):
        api_dict = self.api_dict
        key = self.key
        html_replace_dict = self.html_replace_dict

        class WorkerWebRequestHandler(WebRequestHandler):
            def do_GET(self):
                self._set_headers()

                if self.is_in_path_list([key, 'worker_api']):
                    worker_dict = {
                        'id': api_dict['id'],
                        'name': api_dict['name'],
                        'groups': api_dict['groups'],
                        'health': api_dict['health'],
                        'status': api_dict['status'],
                        'clock': api_dict['clock'],
                    }
                    self.wfile.write(json.dumps(worker_dict, indent=4).encode("utf8"))

                elif self.is_in_path_list([key, 'job_api']):
                    if len(self.path_list) == 3:
                        job_key = self.path_list[2]
                    elif len(self.path_list) == 2:
                        job_key = self.path_list[1]
                    else:
                        job_key = None

                    if job_key:
                        try:
                            job_dict = {
                                'job_detail': api_dict['job_details'][job_key]
                            }
                            self.wfile.write(json.dumps(job_dict, indent=4).encode("utf8"))
                        except KeyError:
                            pass

                elif self.served_css_js():
                    pass

                elif self.is_in_path_list([key]):
                    self.wfile.write(html_encode_file('worker_index.html', replace_dict=html_replace_dict))

                else:
                    self.not_found()

        httpd = HTTPServer((self.address, self.port), WorkerWebRequestHandler)
        httpd.serve_forever()



class Timer:
    def __init__(self):
        self.start_time: float = 0.0
        self.stop_time: float = 0.0

        self.average: float = 0.0
        self.current: float = 0.0
        self.fastest: float = 0.0
        self.previous: float = 0.0
        self.slowest: float = 0.0

    def start(self):
        pass

    def stop(self):
        pass

    def as_dict(self):
        return {
            'average': self.average,
            'current': self.current,
            'fastest': self.fastest,
            'previous': self.previous,
            'slowest': self.slowest,
        }



import threading

from robit.core.clock import Clock
from robit.core.health import Health
from robit.core.id import Id
from robit.job import Job
from robit.core.name import Name
from robit.core.status import Status


class Group:
    def __init__(
            self,
            name: str = 'default',
            utc_offset: int = 0
    ):
        self.id = Id()
        self.name = Name(name)
        self.health = Health()
        self.status = Status()
        self.clock = Clock(utc_offset=utc_offset)

        self.job_list = list()

        self.thread = threading.Thread(target=self.run_job_list)
        self.thread.daemon = True

    def add_job(self, name: str, method, **kwargs):
        self.job_list.append(Job(name=name, method=method, utc_offset=self.clock.utc_offset, **kwargs))

    def calculate_jobs_to_list(self):
        job_list = list()
        self.health.reset()

        for job in self.job_list:
            job_list.append(job.as_dict())
            self.health.average(job.health.percentage)

        return job_list

    def job_list_as_dict_full(self):
        job_dict_full = dict()
        for job in self.job_list:
            job_dict_full[job.id.__str__()] = job.as_dict_full()
        return job_dict_full

    def run_job_list(self):
        while True:
            for job in self.job_list:
                job.run()

    def restart(self):
        pass

    def start(self):
        self.thread.start()

    def stop(self):
        pass

    def as_dict(self):
        return {
            'id': self.id.__str__(),
            'name': self.name.__str__(),
            'health': self.health.__str__(),
            'jobs': self.calculate_jobs_to_list(),
            'status': self.status.__str__(),
        }


import logging
from time import sleep

from robit.core.clock import Clock
from robit.core.counter import Counter
from robit.core.health import Health
from robit.core.id import Id
from robit.core.log import Log
from robit.core.status import Status


class Job:
    def __init__(
            self,
            name: str,
            method,
            utc_offset: int = 0,
            **kwargs
    ):
        self.id = Id()

        self.name = name

        self.method = method

        if 'method_kwargs' in kwargs:
            self.method_kwargs = kwargs['method_kwargs']
        else:
            self.method_kwargs = None

        if 'cron' in kwargs:
            self.clock = Clock(cron=kwargs['cron'], utc_offset=utc_offset)
        else:
            self.clock = Clock(utc_offset=utc_offset)

        self.status = Status()

        self.success_count = Counter()
        self.failed_count = Counter()
        self.failed_log = Log(max_messages=20)

        self.health = Health()

        self.result_log = Log(max_messages=200)

    @property
    def method_verbose(self):
        if self.method_kwargs:
            return f'{self.method.__name__}(kwargs={self.method_kwargs})'
        else:
            return f'{ self.method.__name__ }()'

    def run(self):
        if self.clock.is_past_next_run_datetime():
            logging.warning(f'Starting: Job "{self.name}"')
            self.status.set('run')
            self.clock.start_timer()
            try:
                if self.method_kwargs:
                    method_result = self.method(**self.method_kwargs)
                else:
                    method_result = self.method()
                self.clock.stop_timer()
                logging.warning(f'Success: Job "{self.name}" ran correctly')
                self.success_count.increase()
                self.health.add_positive()
                if method_result:
                    self.result_log.add_message(str(method_result))
            except Exception as e:
                self.status.set('error')
                failed_message = f'Failed on Exception: {e}'
                logging.warning(failed_message)
                self.failed_log.add_message(failed_message)
                self.failed_count.increase()
                self.health.add_negative()
        else:
            if self.status.value != 'error':
                self.status.set('queued')
            sleep(1)

    def as_dict(self):
        return {
            'id': self.id.__str__(),
            'name': self.name.__str__(),
            'status': self.status.__str__(),
            'next_run_datetime': self.clock.next_run_datetime_verbose,
            'success_count': self.success_count.total,
            'health': self.health.__str__(),
            'failed_count': self.failed_count.total,
        }

    def as_dict_full(self):
        return {
            'id': self.id.__str__(),
            'name': self.name.__str__(),
            'method': self.method_verbose,
            'status': self.status.__str__(),
            'result_log': self.result_log.message_list,
            'clock': self.clock.as_dict(),
            'success_count': self.success_count.total,
            'health': self.health.__str__(),
            'failed_count': self.failed_count.total,
            'failed_log': self.failed_log.message_list,
        }



import unittest

from robit import Worker, Monitor


class TestCore(unittest.TestCase):
    def test_worker(self):
        try:
            wo = Worker('Testy McWorkerson')
            self.assertTrue(True)
        except:
            self.assertTrue(False)

    def test_monitor(self):
        try:
            mo = Monitor('Watchy McTesterson')
            self.assertTrue(True)
        except:
            self.assertTrue(False)





from time import sleep

from robit.core.clock import Clock
from robit.core.web_client import post_worker_data_to_monitor
from robit.job.group import Group
from robit.core.health import Health
from robit.core.id import Id
from robit.core.name import Name
from robit.core.status import Status
from robit.worker.web_server import WorkerWebServer


class Worker:
    def __init__(
            self,
            name: str,
            web_server: bool = True,
            web_server_address: str = '127.0.0.1',
            web_server_port: int = 8000,
            key: str = None,
            monitor_address: str = None,
            monitor_port: int = 8200,
            monitor_key: str = None,
            utc_offset: int = 0
    ):
        self.id = Id()
        self.name = Name(name)
        self.clock = Clock(utc_offset=utc_offset)
        self.health = Health()
        self.status = Status()

        if web_server:
            self.web_server = WorkerWebServer(
                address=web_server_address,
                port=web_server_port,
                key=key,
                html_replace_dict={'||title||': self.name.__str__()}
            )
        else:
            self.web_server = None

        self.monitor_address = monitor_address
        self.monitor_port = monitor_port
        self.monitor_key = monitor_key

        self.group_dict = dict()

    def add_job(self, name, method, group='Default', **kwargs):
        if group not in self.group_dict:
            self.group_dict[group] = Group(name=group, utc_offset=self.clock.utc_offset)

        self.group_dict[group].add_job(name, method, **kwargs)

    def as_dict(self):
        return {
            'id': self.id.__str__(),
            'name': self.name.__str__(),
            'groups': self.calculate_groups_to_list(),
            'health': self.health.__str__(),
            'status': self.status.__str__(),
            'clock': self.clock.as_dict(),
            'job_details': self.job_detail_dict()
        }

    def as_dict_to_monitor(self):
        return {
            'id': self.id.__str__(),
            'name': self.name.__str__(),
            'health': self.health.__str__(),
            'clock': self.clock.as_dict(),
        }

    def calculate_groups_to_list(self):
        group_list = list()
        self.health.reset()

        for group in self.group_dict.values():
            group_list.append(group.as_dict())
            self.health.average(group.health.percentage)

        return group_list

    def job_detail_dict(self):
        job_detail_dict = dict()

        for group in self.group_dict.values():
            job_detail_dict = {**job_detail_dict, **group.job_list_as_dict_full()}

        return job_detail_dict

    def restart(self):
        pass

    def start_all_groups(self):
        for group in self.group_dict.values():
            group.start()

    def start(self):
        if self.web_server:
            self.web_server.start()

        self.start_all_groups()

        while True:
            if self.web_server:
                self.web_server.update_api_dict(self.as_dict())
            if self.monitor_address:
                post_worker_data_to_monitor(self.monitor_address, self.monitor_key, self.as_dict_to_monitor())
            sleep(1)

    def stop(self):
        pass



from http.server import BaseHTTPRequestHandler
from pathlib import Path
import threading


def html_encode_file(name, directory: str = 'html', replace_dict: dict = None):
    html = Path(Path(__file__).parent.parent.resolve(), directory, name).read_text()

    if replace_dict:
        html_str = str(html)

        for key, val in replace_dict.items():
            html_str = html_str.replace(key, val)

        return html_str.encode("utf8")

    else:
        return html.encode("utf8")


class WebRequestHandler(BaseHTTPRequestHandler):
    def not_found(self):
        self.wfile.write('Nothing to See Here'.encode("utf8"))

    @property
    def path_list(self):
        temp_path_list = self.path.split('/')
        path_list = list()

        for path in temp_path_list:
            if len(path) > 0:
                path_list.append(path)

        return path_list

    def is_in_path_list(self, path_list: list):
        if len(path_list) >= 1:
            if path_list[0] is None:
                del path_list[0]

        if len(path_list) <= len(self.path_list):
            for i in range(len(path_list)):
                if path_list[i] != self.path_list[i]:
                    return False
            else:
                return True
        else:
            return False

    def served_css_js(self):
        if 1 < len(self.path.split('.')) < 3:
            if self.path[1:].split('.')[1] in ('js', 'css', 'html'):
                if len(self.path.split('/')) > 2:
                    file_name = self.path.split('/')[2]
                else:
                    file_name = self.path[1:]

                self.wfile.write(html_encode_file(file_name))

                return True

        return False

    def _set_headers(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_GET(self):
        pass

    def do_HEAD(self):
        self._set_headers()

    def do_POST(self):
        self._set_headers()
        self.not_found()


class WebServer:
    def __init__(self, address='localhost', port=8000, key=None, html_replace_dict=None):
        self.api_dict = dict()
        self.post_dict = dict()

        self.address = address
        self.port = int(port)
        self.key = key

        self.html_replace_dict = html_replace_dict

        self.thread = threading.Thread(target=self.httpd_serve)
        self.thread.daemon = True

    # Override this Function to customize the webserver
    def httpd_serve(self):
        pass

    def restart(self):
        pass

    def start(self):
        self.thread.start()
        href_link = f'http://{self.address}:{self.port}'

        if self.key:
            href_link += f'/{self.key}/'

        print(f'Starting httpd server at {href_link}')

    def stop(self):
        pass

    def update_api_dict(self, update_dict: dict):
        for key, val in update_dict.items():
            self.api_dict[key] = val



import ast
import json
import urllib.parse
from urllib import parse
from http.server import HTTPServer

from robit.core.web_server import WebServer, WebRequestHandler, html_encode_file


class MonitorWebServer(WebServer):
    def httpd_serve(self):
        api_dict = self.api_dict
        post_dict = self.post_dict
        key = self.key
        html_replace_dict = self.html_replace_dict

        class MonitorWebRequestHandler(WebRequestHandler):
            def do_GET(self):
                self._set_headers()

                print(self.path)

                if self.is_in_path_list([key, 'monitor_api']):
                    self.wfile.write(json.dumps(api_dict, indent=4).encode("utf8"))

                elif self.is_in_path_list([key, 'worker_update']):
                    self.wfile.write('Try Posting to This Path'.encode("utf8"))

                elif self.served_css_js():
                    pass

                elif self.is_in_path_list([key]):
                    self.wfile.write(html_encode_file('monitor_index.html', replace_dict=html_replace_dict))

                else:
                    self.not_found()

            def do_POST(self):
                if self.is_in_path_list([key, 'worker_update']):
                    content_length = int(self.headers['Content-Length'])
                    post_data = self.rfile.read(content_length)
                    post_data_dict = ast.literal_eval(urllib.parse.unquote_plus(str(post_data)[2:-1]))
                    post_dict['worker_dict'][post_data_dict['id']] = post_data_dict
                    # print(f'{post_dict = }')

        httpd = HTTPServer((self.address, self.port), MonitorWebRequestHandler)
        httpd.serve_forever()



from time import sleep

from robit.core.clock import Clock
from robit.core.health import Health
from robit.core.id import Id
from robit.core.name import Name
from robit.core.status import Status
from robit.monitor.web_server import MonitorWebServer


class Monitor:
    def __init__(
            self,
            name: str,
            web_server: bool = True,
            web_server_address: str = '127.0.0.1',
            web_server_port: int = 8200,
            key: str = None,
            utc_offset: int = 0,
    ):
        self.id = Id()
        self.name = Name(name)
        self.clock = Clock(utc_offset=utc_offset)
        self.health = Health()
        self.status = Status()

        if web_server:
            self.web_server = MonitorWebServer(
                address=web_server_address,
                port=web_server_port,
                key=key,
                html_replace_dict={'||title||': self.name.__str__()}
            )
            self.web_server.post_dict['worker_dict'] = dict()
        else:
            self.web_server = None

        self.worker_dict = self.web_server.post_dict['worker_dict']
        # self.monitor_dict = dict()

    def as_dict(self):
        return {
            'id': self.id.__str__(),
            'name': self.name.__str__(),
            'health': self.health.__str__(),
            'status': self.status.__str__(),
            'clock': self.clock.as_dict(),
            'workers': self.calculate_workers_to_list(),
        }

    def calculate_workers_to_list(self):
        worker_list = list()
        self.health.reset()

        for worker in self.worker_dict.values():
            # print(worker)
            worker_list.append(worker)
            self.health.average(float(worker['health']) * 0.01)

        return worker_list

    def restart(self):
        pass

    def start(self):
        if self.web_server:
            self.web_server.start()

        while True:
            if self.web_server:
                self.web_server.update_api_dict(self.as_dict())
                # print(self.worker_dict)
                sleep(1)


    def stop(self):
        pass


class Counter:
    def __init__(self):
        self.total = 0

    def increase(self, count: int = 1):
        self.total += count

    def decrease(self, count: int = 1):
        self.total -= count

    def as_dict(self):
        return {
            'count': self.total
        }


class Health:
    def __init__(self, max_count: int = 500):
        self.max_count: int = max_count
        self.percentage: float = 1.00
        self.positive_count: int = 0
        self.negative_count: int = 0

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.percentage_verbose

    def average(self, value: float):
        if self.percentage == 0.0:
            self.percentage = value
        else:
            self.percentage = (self.percentage + value) / 2

    def calculate(self):
        percentage_value = self.positive_count / (self.positive_count + self.negative_count)
        self.set_percentage(percentage_value)

    def add_negative(self, count: int = 1):
        self.negative_count += count

        if self.negative_count > self.max_count:
            self.negative_count = self.max_count

            if self.positive_count > 0:
                self.positive_count -= count

        self.calculate()

    def add_positive(self, count: int = 1):
        self.positive_count += count

        if self.positive_count > self.max_count:
            self.positive_count = self.max_count

            if self.negative_count > 0:
                self.negative_count -= count

        self.calculate()

    @property
    def percentage_verbose(self):
        return f'{self.percentage * 100:,.2f}'

    def reset(self):
        self.percentage = 0.0

    def set_percentage(self, value: float):
        if value > 100.0:
            self.percentage = 100.0
        elif value < 0.0:
            self.percentage = 0.0
        else:
            self.percentage = value

    def as_dict(self):
        return {
            'percentage': self.percentage,
        }


from robit.core.clock import Clock


class Log:
    def __init__(
            self,
            max_messages: int = 5,
            utc_offset: int = 0,
    ):
        self.max_messages = max_messages
        self.message_list = list()

        self.clock = Clock(utc_offset=utc_offset)

    def add_message(self, message: str):
        self.message_list.insert(0, f'[{self.clock.now_tz_verbose}] {message}')
        self.trim()

    def trim(self):
        if len(self.message_list) > self.max_messages:
            del self.message_list[self.max_messages:]

    def as_dict(self):
        return {
            'log': self.message_list
        }
